@@ -1,6 +1,7 @@
 /*
     commands.c: Commands sent to the card
-    Copyright (C) 2003-2004   Ludovic Rousseau
+    Copyright (C) 2003-2010   Ludovic Rousseau
+    Copyright (C) 2005 Martin Paljak
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -18,7 +19,7 @@
 */
 
 /*
- * $Id: commands.c 3318 2009-02-15 15:55:30Z rousseau $
+ * $Id: commands.c 4902 2010-04-29 07:58:03Z rousseau $
  */
 
 #include <string.h>
@@ -49,7 +50,6 @@
 
 #define max( a, b )   ( ( ( a ) > ( b ) ) ? ( a ) : ( b ) )
 #define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
-#define IFD_ERROR_INSUFFICIENT_BUFFER 700
 
 /* internal functions */
 static RESPONSECODE CmdXfrBlockAPDU_extended(unsigned int reader_index,
@@ -171,6 +171,28 @@ RESPONSECODE CmdPowerOn(unsigned int reader_index, unsigned int * nlength,
 
 	if (ccid_descriptor->dwFeatures & CCID_CLASS_AUTO_VOLTAGE)
 		voltage = 0;	/* automatic voltage selection */
+	else
+	{
+		int bVoltageSupport = ccid_descriptor->bVoltageSupport;
+
+		if ((1 == voltage) && !(bVoltageSupport & 1))
+		{
+			DEBUG_INFO("5V requested but not support by reader");
+			voltage = 2;	/* 3V */
+		}
+
+		if ((2 == voltage) && !(bVoltageSupport & 2))
+		{
+			DEBUG_INFO("3V requested but not support by reader");
+			voltage = 3;	/* 1.8V */
+		}
+
+		if ((3 == voltage) && !(bVoltageSupport & 4))
+		{
+			DEBUG_INFO("1.8V requested but not support by reader");
+			voltage = 0;	/* auto */
+		}
+	}
 
 again:
 	cmd[0] = 0x62; /* IccPowerOn */
@@ -302,7 +324,8 @@ RESPONSECODE SecurePINVerify(unsigned int reader_index,
 
 #ifdef BOGUS_PINPAD_FIRMWARE
 	/* bug circumvention for the GemPC Pinpad */
-	if (GEMPCPINPAD == ccid_descriptor->readerID)
+	if ((GEMPCPINPAD == ccid_descriptor->readerID)
+		|| (VEGAALPHA == ccid_descriptor->readerID))
 	{
 		/* the firmware reject the cases: 00h No string and FFh default
 		 * CCID message. The only value supported is 01h (display 1 message) */
@@ -323,6 +346,34 @@ RESPONSECODE SecurePINVerify(unsigned int reader_index,
 		}
 
 	}
+
+	if ((DELLSCRK == ccid_descriptor->readerID)
+		|| (DELLSK == ccid_descriptor->readerID))
+	{
+		/* the firmware rejects the cases: 01h-FEh and FFh default
+		 * CCID message. The only value supported is 00h (no message) */
+		if (0x00 != TxBuffer[8])
+		{
+			DEBUG_INFO2("Correct bNumberMessage for Dell keyboard (was %d)",
+				TxBuffer[8]);
+			TxBuffer[8] = 0x00;
+		}
+
+		/* avoid the command rejection because the Enter key is still
+		 * pressed. Wait a bit for the key to be released */
+		(void)usleep(250*1000);
+	}
+
+	if (DELLSK == ccid_descriptor->readerID)
+	{
+		/* the 2 bytes of wPINMaxExtraDigit are reversed */
+		int tmp;
+
+		tmp = TxBuffer[6];
+		TxBuffer[6] = TxBuffer[5];
+		TxBuffer[5] = tmp;
+		DEBUG_INFO("Correcting wPINMaxExtraDigit for Dell keyboard");
+	}
 #endif
 
 	/* T=1 Protocol Management for a TPDU reader */
@@ -351,7 +402,7 @@ RESPONSECODE SecurePINVerify(unsigned int reader_index,
 			sdata, 3);
 	}
 
-	/* Build a CCID block from a PC/SC V2.1.2 Part 10 block */
+	/* Build a CCID block from a PC/SC V2.02.05 Part 10 block */
 	for (a = 11, b = 0; b < TxLength; b++)
 	{
 		if (1 == b) /* bTimeOut2 field */
@@ -399,17 +450,21 @@ RESPONSECODE SecurePINVerify(unsigned int reader_index,
 	ccid_descriptor -> readTimeout = max(30, TxBuffer[0]+10);	/* at least 30 seconds */
 
 	if (WritePort(reader_index, a, cmd) != STATUS_SUCCESS)
-		return IFD_COMMUNICATION_ERROR;
+	{
+		ret = IFD_COMMUNICATION_ERROR;
+		goto end;
+	}
 
 	ret = CCID_Receive(reader_index, RxLength, RxBuffer, NULL);
 
 	/* T=1 Protocol Management for a TPDU reader */
-	if ((IFD_SUCCESS == ret)
-		&& (SCARD_PROTOCOL_T1 == ccid_descriptor->cardProtocol)
+	if ((SCARD_PROTOCOL_T1 == ccid_descriptor->cardProtocol)
 		&& (CCID_CLASS_TPDU == (ccid_descriptor->dwFeatures & CCID_CLASS_EXCHANGE_MASK)))
 	{
 		/* timeout and cancel cases are faked by CCID_Receive() */
-		if (2 == *RxLength)
+		if ((2 == *RxLength)
+			/* the CCID command is rejected or failed */
+		   || (IFD_SUCCESS != ret))
 		{
 			/* Decrement the sequence numbers since no TPDU was sent */
 			get_ccid_slot(reader_index)->t1.ns ^= 1;
@@ -424,6 +479,7 @@ RESPONSECODE SecurePINVerify(unsigned int reader_index,
 		}
 	}
 
+end:
 	ccid_descriptor -> readTimeout = old_read_timeout;
 	return ret;
 } /* SecurePINVerify */
@@ -475,8 +531,9 @@ RESPONSECODE SecurePINModify(unsigned int reader_index,
 		return IFD_NOT_SUPPORTED;
 	}
 
-	/* Make sure in the beginning if bNumberMessage is valid or not */
-	if (TxBuffer[11] > 3)
+	/* Make sure in the beginning if bNumberMessage is valid or not.
+	 * 0xFF is the default value. */
+	if ((TxBuffer[11] > 3) && (TxBuffer[11] != 0xFF))
 	{
 		DEBUG_INFO2("Wrong bNumberMessage: %d", TxBuffer[11]);
 		return IFD_NOT_SUPPORTED;
@@ -515,7 +572,8 @@ RESPONSECODE SecurePINModify(unsigned int reader_index,
 	}
 
 	/* bug circumvention for the GemPC Pinpad */
-	if (GEMPCPINPAD == ccid_descriptor->readerID)
+	if ((GEMPCPINPAD == ccid_descriptor->readerID)
+		|| (VEGAALPHA == ccid_descriptor->readerID))
 	{
 		/* The reader does not support, and actively reject, "max size reached"
 		 * and "timeout occured" validation conditions */
@@ -564,7 +622,7 @@ RESPONSECODE SecurePINModify(unsigned int reader_index,
 			sdata, 3);
 	}
 
-	/* Build a CCID block from a PC/SC V2.1.2 Part 10 block */
+	/* Build a CCID block from a PC/SC V2.02.05 Part 10 block */
 
 	/* Do adjustments as needed - CCID spec is not exact with some
 	 * details in the format of the structure, per-reader adaptions
@@ -608,7 +666,8 @@ RESPONSECODE SecurePINModify(unsigned int reader_index,
 		cmd[21] = 0x00; /* set bNumberMessages to 0 */
 	}
 
-	if (GEMPCPINPAD == ccid_descriptor->readerID)
+	if ((GEMPCPINPAD == ccid_descriptor->readerID)
+		|| (VEGAALPHA == ccid_descriptor->readerID))
 		cmd[21] = bNumberMessages;	/* restore the real value */
 #endif
 
@@ -619,17 +678,21 @@ RESPONSECODE SecurePINModify(unsigned int reader_index,
 	ccid_descriptor -> readTimeout = max(30, TxBuffer[0]+10);	/* at least 30 seconds */
 
 	if (WritePort(reader_index, a, cmd) != STATUS_SUCCESS)
- 		return IFD_COMMUNICATION_ERROR;
+	{
+ 		ret = IFD_COMMUNICATION_ERROR;
+		goto end;
+	}
 
  	ret = CCID_Receive(reader_index, RxLength, RxBuffer, NULL);
 
 	/* T=1 Protocol Management for a TPDU reader */
-	if ((IFD_SUCCESS == ret)
-		&& (SCARD_PROTOCOL_T1 == ccid_descriptor->cardProtocol)
+	if ((SCARD_PROTOCOL_T1 == ccid_descriptor->cardProtocol)
 		&& (CCID_CLASS_TPDU == (ccid_descriptor->dwFeatures & CCID_CLASS_EXCHANGE_MASK)))
 	{
 		/* timeout and cancel cases are faked by CCID_Receive() */
-		if (2 == *RxLength)
+		if ((2 == *RxLength)
+			/* the CCID command is rejected or failed */
+			|| (IFD_SUCCESS != ret))
 		{
 			/* Decrement the sequence numbers since no TPDU was sent */
 			get_ccid_slot(reader_index)->t1.ns ^= 1;
@@ -644,6 +707,7 @@ RESPONSECODE SecurePINModify(unsigned int reader_index,
 		}
 	}
 
+end:
 	ccid_descriptor -> readTimeout = old_read_timeout;
 	return ret;
 } /* SecurePINModify */
@@ -662,19 +726,27 @@ RESPONSECODE CmdEscape(unsigned int reader_index,
 	status_t res;
 	unsigned int length_in, length_out;
 	RESPONSECODE return_value = IFD_SUCCESS;
+	int old_read_timeout;
 	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
 
+	old_read_timeout = ccid_descriptor -> readTimeout;
+	ccid_descriptor -> readTimeout = 30;	/* 30 seconds */
+
 again:
 	/* allocate buffers */
 	length_in = 10 + TxLength;
 	if (NULL == (cmd_in = malloc(length_in)))
-		return IFD_COMMUNICATION_ERROR;
+	{
+		return_value = IFD_COMMUNICATION_ERROR;
+		goto end;
+	}
 
 	length_out = 10 + *RxLength;
 	if (NULL == (cmd_out = malloc(length_out)))
 	{
 		free(cmd_in);
-		return IFD_COMMUNICATION_ERROR;
+		return_value = IFD_COMMUNICATION_ERROR;
+		goto end;
 	}
 
 	cmd_in[0] = 0x6B; /* PC_to_RDR_Escape */
@@ -691,7 +763,8 @@ again:
 	if (res != STATUS_SUCCESS)
 	{
 		free(cmd_out);
-		return IFD_COMMUNICATION_ERROR;
+		return_value = IFD_COMMUNICATION_ERROR;
+		goto end;
 	}
 
 	res = ReadPort(reader_index, &length_out, cmd_out);
@@ -709,13 +782,15 @@ again:
 	if (res != STATUS_SUCCESS)
 	{
 		free(cmd_out);
-		return IFD_COMMUNICATION_ERROR;
+		return_value = IFD_COMMUNICATION_ERROR;
+		goto end;
 	}
 
 	if (length_out < STATUS_OFFSET+1)
 	{
 		DEBUG_CRITICAL2("Not enough data received: %d bytes", length_out);
-		return IFD_COMMUNICATION_ERROR;
+		return_value = IFD_COMMUNICATION_ERROR;
+		goto end;
 	}
 
 	if (cmd_out[STATUS_OFFSET] & CCID_COMMAND_FAILED)
@@ -733,6 +808,8 @@ again:
 
 	free(cmd_out);
 
+end:
+	ccid_descriptor -> readTimeout = old_read_timeout;
 	return return_value;
 } /* Escape */
 
@@ -849,7 +926,7 @@ RESPONSECODE CmdGetSlotStatus(unsigned int reader_index, unsigned char buffer[])
 
 again_status:
 		/* SlotStatus */
-		r = ControlUSB(reader_index, 0xA1, 0xA0, 0, status, sizeof(status)); 
+		r = ControlUSB(reader_index, 0xA1, 0xA0, 0, status, sizeof(status));
 
 		/* we got an error? */
 		if (r < 0)
@@ -942,20 +1019,12 @@ again_status:
 		return IFD_COMMUNICATION_ERROR;
 	}
 
-	if (buffer[STATUS_OFFSET] & CCID_COMMAND_FAILED)
-	{
-#ifdef O2MICRO_OZ776_PATCH
-		/* the O2MICRO OZ 776 reader sends card absent or mute errors
-		 * when no card is inserted */
-		if (! (((OZ776 == ccid_descriptor->readerID)
-			|| (OZ776_7772 == ccid_descriptor->readerID))
-			&& (buffer[ERROR_OFFSET] == 0xFE)))
-#endif
-		ccid_error(buffer[ERROR_OFFSET], __FILE__, __LINE__, __FUNCTION__);    /* bError */
-
+	if ((buffer[STATUS_OFFSET] & CCID_COMMAND_FAILED)
 		/* card absent or mute is not an communication error */
-		if (buffer[ERROR_OFFSET] != 0xFE)
-			return_value = IFD_COMMUNICATION_ERROR;
+		&& (buffer[ERROR_OFFSET] != 0xFE))
+	{
+		return_value = IFD_COMMUNICATION_ERROR;
+		ccid_error(buffer[ERROR_OFFSET], __FILE__, __LINE__, __FUNCTION__);    /* bError */
 	}
 
 	return return_value;
@@ -1037,7 +1106,8 @@ RESPONSECODE CCID_Transmit(unsigned int reader_index, unsigned int tx_length,
 		int r;
 
 		/* Xfr Block */
-		r = ControlUSB(reader_index, 0x21, 0x65, 0, tx_buffer, tx_length);
+		r = ControlUSB(reader_index, 0x21, 0x65, 0,
+			(unsigned char *)tx_buffer, tx_length);
 
 		/* we got an error? */
 		if (r < 0)
@@ -1059,8 +1129,8 @@ RESPONSECODE CCID_Transmit(unsigned int reader_index, unsigned int tx_length,
 
 		/* Xfr Block */
 		DEBUG_COMM2("chain parameter: %d", rx_length);
-		r = ControlUSB(reader_index, 0x21, 0x65, rx_length << 8, tx_buffer,
-			tx_length);
+		r = ControlUSB(reader_index, 0x21, 0x65, rx_length << 8,
+			(unsigned char *)tx_buffer, tx_length);
 
 		/* we got an error? */
 		if (r < 0)
@@ -1130,6 +1200,9 @@ RESPONSECODE CCID_Receive(unsigned int reader_index, unsigned int *rx_length,
 			return IFD_COMMUNICATION_ERROR;
 		}
 
+		/* we need to store returned value */
+		*rx_length = r;
+
 		return IFD_SUCCESS;
 	}
 
@@ -1436,7 +1509,10 @@ receive_next_block:
 		 * next PC_to_RDR_XfrBlock command */
 		case 0x10:
 			/* send a nul block */
-			return_value = CCID_Transmit(reader_index, 0, NULL, 0, 0);
+			/* set wLevelParameter to 0010h: empty abData field,
+			 * continuation of response APDU is
+			 * expected in the next RDR_to_PC_DataBlock. */
+			return_value = CCID_Transmit(reader_index, 0, NULL, 0x10, 0);
 			if (return_value != IFD_SUCCESS)
 				return return_value;
 
@@ -1728,6 +1804,7 @@ static RESPONSECODE CmdXfrBlockCHAR_T0(unsigned int reader_index,
 	unsigned char tmp_buf[512];
 	unsigned int exp_len, in_len;
 	unsigned char ins, *in_buf;
+	unsigned char backup_len = *rcv_len;
 	RESPONSECODE return_value = IFD_SUCCESS;
 	_ccid_descriptor *ccid_descriptor = get_ccid_descriptor(reader_index);
 
@@ -1737,6 +1814,12 @@ static RESPONSECODE CmdXfrBlockCHAR_T0(unsigned int reader_index,
 	{
 		unsigned char pcbuffer[SIZE_GET_SLOT_STATUS];
 
+		/* length is on 16-bits only
+		 * if a size > 0x1000 is used then usb_control_msg() fails with
+		 * "Invalid argument" */
+		if (*rcv_len > 0x1000)
+			*rcv_len = 0x1000;
+
 		/* Command to send to the smart card (must be 5 bytes) */
 		memset(cmd, 0, sizeof(cmd));
 		if (snd_len == 4)
@@ -1762,40 +1845,48 @@ static RESPONSECODE CmdXfrBlockCHAR_T0(unsigned int reader_index,
 		if (return_value != IFD_SUCCESS)
 			return return_value;
 
-		if ((0x20 != pcbuffer[0]) && (snd_len > 0))
+		if (0x10 == pcbuffer[0])
 		{
-			/* continue sending the APDU */
-			return_value = CCID_Transmit(reader_index, snd_len, snd_buf, 0, 0);
-			if (return_value != IFD_SUCCESS)
-				return return_value;
-		}
-		else
-		{
-			if ((0x20 == pcbuffer[0]) && (*rcv_len > 2))
-				/* we will just read SW1-SW2 */
-				*rcv_len = 2;
-
-			return_value = CCID_Receive(reader_index, rcv_len, rcv_buf, NULL);
-			if (return_value != IFD_SUCCESS)
-				DEBUG_CRITICAL("CCID_Receive failed");
-
-			return return_value;
+			if (snd_len > 0)
+			{
+				/* continue sending the APDU */
+				return_value = CCID_Transmit(reader_index, snd_len, snd_buf,
+					0, 0);
+				if (return_value != IFD_SUCCESS)
+					return return_value;
+			}
+			else
+			{
+				/* read apdu data */
+				return_value = CCID_Receive(reader_index, rcv_len, rcv_buf,
+						NULL);
+				if (return_value != IFD_SUCCESS)
+					return return_value;
+			}
 		}
 
-		/* wait for ready */
 		return_value = CmdGetSlotStatus(reader_index, pcbuffer);
 		if (return_value != IFD_SUCCESS)
 			return return_value;
 
-		if ((0x20 == pcbuffer[0]) && (*rcv_len > 2))
-			/* we will just read SW1-SW2 */
+		/* SW1-SW2 available */
+		if (0x20 == pcbuffer[0])
+		{
+			/* backup apdu data length */
+			/* if no data recieved before - backup length must be zero */
+			backup_len = (backup_len == *rcv_len) ? 0 : *rcv_len;
+
+			/* wait for 2 bytes (SW1-SW2) */
 			*rcv_len = 2;
 
-		/* read SW1-SW2 */
-		return_value = CCID_Receive(reader_index, rcv_len, rcv_buf, NULL);
-		if (return_value != IFD_SUCCESS)
-			DEBUG_CRITICAL("CCID_Receive failed");
+			return_value = CCID_Receive(reader_index, rcv_len,
+				rcv_buf + backup_len, NULL);
+			if (return_value != IFD_SUCCESS)
+				DEBUG_CRITICAL("CCID_Receive failed");
 
+			/* restore recieved length */
+			*rcv_len += backup_len;
+		}
 		return return_value;
 	}
 
